from __future__ import absolute_import

from .contract import LazyContract
from .properties import (ObjectProperty, ListProperty, SetProperty,
                         DictProperty, StringProperty, IntegerProperty,
                         FloatProperty, BooleanProperty)


def test_basic_properties():
    class TestContract(LazyContract):
        a = StringProperty()
        b = IntegerProperty()
        c = FloatProperty()
        d = BooleanProperty()

    expected = dict(a='foo', b=1, c=2.3, d=False)
    t1 = TestContract(a='foo', b=1, c=2.3, d=False)
    t2 = TestContract(a='foo', b='1', c='2.3', d='false')
    assert t1.to_dict() == expected
    assert t2.to_dict() == expected


def test_object_property():
    class TestContract(LazyContract):
        a = StringProperty()
        b = IntegerProperty()
        c = FloatProperty()

    class ObjectPropertyContract(LazyContract):
        a = ObjectProperty(TestContract)
        b = ObjectProperty(TestContract)

    x = ObjectPropertyContract(a=dict(a='foo', b=1, c=2.3),
                               b=TestContract(a='bar', b=4, c=5.6))
    assert x.to_dict() == dict(a=dict(a='foo', b=1, c=2.3),
                               b=dict(a='bar', b=4, c=5.6))


def test_list_property():
    class TestContract(LazyContract):
        a = StringProperty()
        b = IntegerProperty()
        c = FloatProperty()

    class ListPropertyContract(LazyContract):
        a = ListProperty()
        i = ListProperty(IntegerProperty())
        t = ListProperty(ObjectProperty(TestContract))

    x = ListPropertyContract(
        a=[1, 'foo'],
        i=[1, 2, '3'],
        t=[dict(a='foo', b=2, c=3.4), TestContract(a='bar', b=4, c=5.6)])

    expected = dict(a=[1, 'foo'],
                    i=[1, 2, 3],
                    t=[dict(a='foo', b=2, c=3.4),
                       dict(a='bar', b=4, c=5.6)])

    assert x.to_dict() == expected


def test_dict_property():
    class TestContract(LazyContract):
        a = StringProperty()
        b = IntegerProperty()
        c = FloatProperty()

    class DictPropertyContract(LazyContract):
        a = DictProperty()
        i = DictProperty(IntegerProperty())
        t = DictProperty(ObjectProperty(TestContract))

    x = DictPropertyContract(
        a={1: 'foo'},
        i={1: '2'},
        t={'t': TestContract(a='bar', b=4, c=5.6)})

    assert x.to_dict() == dict(a={1: 'foo'},
                               i={1: 2},
                               t={'t': dict(a='bar', b=4, c=5.6)})


def test_set_property():

    class TestContract(LazyContract):
        a = StringProperty()
        b = IntegerProperty()
        c = FloatProperty()

    class SetPropertyContract(LazyContract):
        a = SetProperty()
        i = SetProperty(IntegerProperty())
        t = SetProperty(ObjectProperty(TestContract))

    x = SetPropertyContract(
        a={1, 'foo'},
        i={1, '2'})

    assert x.to_dict() == dict(a={1, 'foo'}, i={1, 2})



from __future__ import absolute_import

import six


class LazyContractError(Exception):
    ''' Catch-all exception type '''

    pass


class LazyContractDeserializationError(LazyContractError):
    ''' Exception deserializing a message into a LazyContract-derived class '''

    FMT = 'failed to deserialize {}.{} from {} due to: {}'


class LazyContractValidationError(LazyContractError):
    ''' Exception validating data within a LazyContract-derived class '''

    INVALID_ATTR_FMT = '{} has no attribute \'{}\''
    NOT_NONE_FMT = '{} {} must not be None'
    REQUIRED_FMT = '{}.{} is required for {}'
    ATTR_TYPE_FMT = '{}.{} value {} is not of type {}'


class LazyProperty(object):
    ''' Base class for descriptors used as properties in a LazyContract.
        Create a sub-class of this to define your own (de-)serialization.
    '''

    _type = type(None)  # defines the type that instances take on

    _default_name = '(anonymous)'  # applies to properties within a container

    def __init__(self, name=None, default=None,
                 required=False, not_none=False, exclude_if_none=True):
        ''' Create a LazyProperty.
            name (string):   the attribute name used for (de-)serialization
            default (_type): default value if not available during deserialization
            required (bool): raise LazyContractValidationError if not provided
            not_none (bool): raise LazyContractValidationError if value is None
            exclude_if_none (bool): don't serialize if value is None
        '''

        if required and default is not None:
            raise LazyContractError('default specified for required property')

        self.name = name or self._default_name
        self.required = required
        self.default = default
        self.not_none = not_none
        self.exclude_if_none = exclude_if_none

    def __get__(self, obj, objtype=None):
        value = obj.__dict__.get(self.name, self.default)
        self.validate(value)
        return value

    def __set__(self, obj, value):
        self.validate(value)
        obj.__dict__[self.name] = value

    def validate(self, obj):
        ''' Validate values whenever they're set.
            Raises LazyContractValidationError.
            Sub-classes that override this probably want to call this method.
        '''

        if obj is None and self.not_none:
            raise LazyContractValidationError(
                    LazyContractValidationError.NOT_NONE_FMT.format(
                            type(self).__name__, self.name))

        if not isinstance(obj, self._type) and obj is not None:
            raise LazyContractValidationError(
                    LazyContractValidationError.ATTR_TYPE_FMT.format(
                            type(self).__name__, self.name, repr(obj), self._type))

    def serialize(self, obj):
        ''' Convert the value to a basic serializable type (ex. str)
            Sub-classes probably need to override this.
        '''

        return obj

    def deserialize(self, obj):
        ''' Convert abasic type to _type.
            Sub-classes probably need to override this.
        '''

        return obj if isinstance(obj, self._type) else self._type(obj)


class LazyContract(object):
    ''' Base class for custom contracts to be serialized and deserialized. '''

    _ignore_undefined_attributes = True

    def __init__(self, _obj=None, **kwargs):
        ''' Deserialize a contract.
            _obj (dict): data to deserialize
            kwargs:      individual attributes to deserialize
        '''

        if _obj is not None and kwargs:
            raise LazyContractError('both _obj and kwargs provided')

        self._properties = dict()
        self._mappings = dict()

        for cls in reversed(self.__class__.__mro__):
            if cls != LazyContract and issubclass(cls, LazyContract):
                self.__discover_properties(_obj or kwargs, cls)

        self._populate_properties(_obj or kwargs)

    def __repr__(self):
        return '{}({})'.format(
            self.__class__.__name__,
            ', '.join('{}={}'.format(name, repr(value))
                      for name, _, value in self.__iter_properties()))

    def __discover_properties(self, obj, cls):
        for name, inst in six.iteritems(cls.__dict__):
            if isinstance(inst, LazyProperty):
                self._properties[name] = inst

                if inst.name == inst._default_name:
                    inst.name = name
                else:
                    self._mappings[inst.name] = name

                if inst.name not in obj and name not in obj:
                    if inst.required:
                        raise LazyContractValidationError(
                                LazyContractValidationError.REQUIRED_FMT.format(
                                        type(self).__name__, inst.name, repr(obj)))

                    if inst.not_none and inst.default is None:
                        raise LazyContractValidationError(
                                LazyContractValidationError.NOT_NONE_FMT.format(
                                        type(self).__name__, inst.name))

    def _populate_properties(self, obj):
        for key, value in six.iteritems(obj):
            if key in self._mappings:
                key = self._mappings[key]

            if key not in self._properties:
                if self._ignore_undefined_attributes:
                    continue
                raise LazyContractValidationError(
                        LazyContractValidationError.INVALID_ATTR_FMT.format(
                                type(self).__name__, key))

            if value is not None:
                try:
                    value = self._properties[key].deserialize(value)
                except Exception as e:
                    raise LazyContractDeserializationError(
                            LazyContractDeserializationError.FMT.format(
                                    type(self).__name__, key, repr(value), e))

            setattr(self, key, value)

    def __iter_properties(self):
        for name, prop in six.iteritems(self._properties):
            yield name, prop, getattr(self, name)

    def to_dict(self):
        ''' Serialize the contract object into a Python dictionary '''

        return {prop.name: prop.serialize(value)
                for name, prop, value in self.__iter_properties()
                if not prop.name.startswith('_') and
                (value is not None or not prop.exclude_if_none)}


class StrictContract(LazyContract):
    ''' Variant of LazyContract does not allow undefined attributes
    '''

    _ignore_undefined_attributes = False


class DynamicContract(LazyContract):
    ''' Variant of LazyContract that deserializes undeclared attributes '''

    def _populate_properties(self, obj):
        contained = {k: v for k, v in six.iteritems(obj) \
                     if k in self._properties or k in self._mappings}
        super(DynamicContract, self)._populate_properties(contained)

        for key, value in six.iteritems(obj):
            if key not in self._properties and key not in self._mappings:
                setattr(self, key, value)



from __future__ import absolute_import

from .contract import LazyContract
from .extra import AliasProperty, UUIDProperty, UUIDStringProperty

import uuid


def test_uuid_property():
    class TestUUIDContract(LazyContract):
        g = UUIDProperty()

    example = '14d0a7b5-33c5-439b-a66b-2d464f4e7d1b'
    expected = dict(g=example)

    assert TestUUIDContract(g=example).to_dict() == expected
    assert TestUUIDContract(g=example.upper()).g == uuid.UUID(example)
    assert type(TestUUIDContract(g=example.upper()).g) == uuid.UUID
    assert TestUUIDContract(g=uuid.UUID(example)).to_dict() == expected


def test_uuid_string_property():
    class TestUUIDStringContract(LazyContract):
        g = UUIDStringProperty()

    example = '14d0a7b5-33c5-439b-a66b-2d464f4e7d1b'
    expected = dict(g=example)

    assert TestUUIDStringContract(g=example).to_dict() == expected
    assert TestUUIDStringContract(g=example.upper()).g == example
    assert type(TestUUIDStringContract(g=example.upper()).g) == type(example)
    assert TestUUIDStringContract(g=uuid.UUID(example)).to_dict() == expected


def test_alias_property():
    class TestAliasContract(LazyContract):
        a = UUIDStringProperty()
        b = AliasProperty('a')

    example = '14d0a7b5-33c5-439b-a66b-2d464f4e7d1b'
    expected = dict(a=example, b=example)

    assert TestAliasContract(a=example).to_dict() == expected

    t = TestAliasContract()
    t.b = example
    assert t.to_dict() == expected



from __future__ import absolute_import

from .contract import LazyProperty, LazyContract

import six


class StringProperty(LazyProperty):

    _type = six.string_types

    def deserialize(self, obj):
        return obj if isinstance(obj, self._type) else str(obj)


class BooleanProperty(LazyProperty):

    _type = bool

    def deserialize(self, obj):
        if isinstance(obj, six.string_types):
            return obj.lower() == 'true'
        else:
            return self._type(obj)


class IntegerProperty(LazyProperty):

    _type = six.integer_types

    def deserialize(self, obj):
        return obj if isinstance(obj, self._type) else int(obj)


class FloatProperty(LazyProperty):

    _type = float


class ObjectProperty(LazyProperty):

    _type = LazyContract

    def __init__(self, kind, *args, **kwargs):
        assert issubclass(kind, LazyContract)
        super(ObjectProperty, self).__init__(*args, **kwargs)
        self._kind = kind

    def serialize(self, obj):
        return obj.to_dict()

    def deserialize(self, obj):
        return obj if isinstance(obj, self._kind) else self._kind(obj)


class ContainerProperty(LazyProperty):

    def __init__(self, lazyproperty=None, *args, **kwargs):
        assert lazyproperty is None or isinstance(lazyproperty, LazyProperty)
        super(ContainerProperty, self).__init__(*args, **kwargs)
        self._property = lazyproperty


class ListProperty(ContainerProperty):

    _type = list

    def serialize(self, obj):
        if self._property is None:
            return obj
        else:
            return [self._property.serialize(e) for e in obj]

    def deserialize(self, obj):
        if self._property is None:
            return obj
        else:
            return [self._property.deserialize(e) for e in obj]


class DictProperty(ContainerProperty):

    _type = dict

    def serialize(self, obj):
        if self._property is None:
            return obj
        else:
            return {k: self._property.serialize(e)
                    for k, e in six.iteritems(obj)}

    def deserialize(self, obj):
        if self._property is None:
            return obj
        else:
            return {k: self._property.deserialize(e)
                    for k, e in six.iteritems(obj)}


class SetProperty(ContainerProperty):

    _type = set

    def serialize(self, obj):
        if self._property is None:
            return obj
        else:
            return {self._property.serialize(e) for e in obj}

    def deserialize(self, obj):
        if self._property is None:
            return obj
        else:
            return {self._property.deserialize(e) for e in obj}



from __future__ import absolute_import

from .contract import LazyProperty

import uuid


class AliasProperty(LazyProperty):

    def __init__(self, aliased_property, *args, **kwargs):
        super(AliasProperty, self).__init__(*args, **kwargs)
        self.aliased_property = aliased_property

    def __get__(self, obj, objtype=None):
        return getattr(obj, self.aliased_property)

    def __set__(self, obj, value):
        setattr(obj, self.aliased_property, value)


class UUIDProperty(LazyProperty):

    _type = uuid.UUID

    def deserialize(self, obj):
        return obj if isinstance(obj, self._type) else uuid.UUID(obj)

    def serialize(self, obj):
        return str(obj).lower()


class UUIDStringProperty(LazyProperty):

    _type = str

    def deserialize(self, obj):
        if isinstance(obj, uuid.UUID):
            return str(obj).lower()
        else:
            return str(uuid.UUID(obj)).lower()

    def serialize(self, obj):
        return obj.lower()



#!/usr/bin/env python

from setuptools import setup

setup(name='lazycontract',
      version='0.9.2',
      description='Python library to define declarative contracts for serialization and deserialization',
      long_description='Please see https://github.com/neilisaac/lazycontract for details',
      author='Neil Isaac',
      author_email='isaac.neil@gmail.com',
      license='MIT',
      platforms=['Any'],
      url='https://github.com/neilisaac/lazycontract',
      packages=['lazycontract'],
      install_requires=['six'],
      tests_require=['pytest', 'tox'],
      classifiers=[
          'Intended Audience :: Developers',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3.2',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Topic :: Software Development :: Libraries']
      )



from __future__ import absolute_import

from .contract import (LazyContract, StrictContract, DynamicContract, LazyContractValidationError,
                       LazyContractDeserializationError)
from .properties import StringProperty, IntegerProperty, FloatProperty


def test_to_dict():
    class TestContract(LazyContract):
        a = StringProperty()

    x = TestContract(a='foo')
    x.y = 4
    s = x.to_dict()
    assert type(s) == dict
    assert s['a'] == 'foo'
    assert len(s) == 1
    assert 'y' not in s


def test_kwargs_deserialization():
    class TestContract(LazyContract):
        a = StringProperty()

    x = TestContract(a='foo')
    assert x.a == 'foo'
    assert repr(x) == 'TestContract(a=\'foo\')'


def test_dict_deserialization():
    class TestContract(LazyContract):
        a = StringProperty()

    x = TestContract({'a': 'foo'})
    assert x.a == 'foo'
    assert repr(x) == 'TestContract(a=\'foo\')'


def test_strict_contract_invalid_deserialize_attribute():
    class TestContract(StrictContract):
        a = StringProperty()
        b = IntegerProperty()
        c = FloatProperty()

    try:
        TestContract({'x': 'foo'})
        assert 'expected LazyContractValidationError' == False
    except LazyContractValidationError as e:
        assert type(e) == LazyContractValidationError
        assert LazyContractValidationError.INVALID_ATTR_FMT.format(
                'TestContract', 'x') in str(e)


def test_required():
    class TestContract(LazyContract):
        a = StringProperty(required=True)

    try:
        TestContract()
        assert 'LazyContractValidationError expected' == False
    except LazyContractValidationError as e:
        assert LazyContractValidationError.REQUIRED_FMT.format(
                'TestContract', 'a', '') in str(e)


def test_not_none():
    class TestContract(LazyContract):
        a = StringProperty(not_none=True)

    try:
        TestContract()
        assert 'LazyContractValidationError expected' == False
    except LazyContractValidationError as e:
        assert LazyContractValidationError.NOT_NONE_FMT.format(
                'TestContract', 'a') in str(e)

    try:
        TestContract(a=None)
        assert 'LazyContractValidationError expected' == False
    except LazyContractValidationError as e:
        assert LazyContractValidationError.NOT_NONE_FMT.format(
                'StringProperty', 'a') in str(e)

    t = TestContract(a='foobar')
    try:
        t.a = None
        assert 'LazyContractValidationError expected' == False
    except LazyContractValidationError as e:
        assert LazyContractValidationError.NOT_NONE_FMT.format(
                'StringProperty', 'a') in str(e)


def test_hidden_property():
    class TestContract(LazyContract):
        _a = StringProperty()

    assert TestContract(_a='foobar').to_dict() == dict()


def test_exclude_if_none():
    class TestContract(LazyContract):
        a = StringProperty(exclude_if_none=False)

    assert TestContract().to_dict() == dict(a=None)


def test_deserialization_error():
    class TestContract(LazyContract):
        a = IntegerProperty()

    reason = None
    try:
        int('3.2')
    except Exception as e:
        reason = e

    try:
        TestContract(a='3.2')
        assert 'expected LazyContractDeserializationError' == False
    except LazyContractDeserializationError as e:
        assert LazyContractDeserializationError.FMT.format(
                'TestContract', 'a', repr('3.2'), reason) in str(e)


def test_serialization_name():
    class TestContract(LazyContract):
        a = StringProperty(name='b')

    t = TestContract(a='a1')
    assert t.to_dict() == dict(b='a1')

    t.a = 'a2'
    assert t.to_dict() == dict(b='a2')

    t = TestContract(b='b1')
    assert t.to_dict() == dict(b='b1')

    t.b = 'b2'
    assert t.to_dict() == dict(b='b2')


def test_inheritence():
    class TestContract1(LazyContract):
        a = StringProperty()

    class TestContract2(TestContract1):
        b = StringProperty()

    t = TestContract2(a='foo', b='bar')
    assert t.to_dict() == dict(a='foo', b='bar')


def test_lazy_contract_ignores_attribute():
    class TestContract(LazyContract):
        a = StringProperty()

    t = TestContract(a='foo', b='bar')
    assert t.to_dict() == dict(a='foo')
    try:
        t.b
        assert 'Expected AttributeError' == False
    except AttributeError:
        pass


def test_dynamic_contract():
    class TestContract(DynamicContract):
        a = StringProperty()

    t = TestContract(a='foo', b='bar')
    assert t.to_dict() == dict(a='foo')
    assert t.b == 'bar'
